/*
$Id: fac.c,v 1.4 2010/12/04 00:47:47 scott Exp $

Description: FASTA file change and clearing utility

$Log: fac.c,v $
Revision 1.4  2010/12/04 00:47:47  scott
Change gap strip option to strip any non-alpha character

Revision 1.3  2010/08/06 01:21:20  scott
Use 'stderr' for printing errors and usage.

Revision 1.2  2010/07/27 21:43:42  scott
Update for calls to Fasta_ routines

History:
 04-23-09 Scott: original revision
 08-31-09 Scott: add clear gap symbols option, use fasta.[ch]

*/

#include <stdio.h>
#include <stdlib.h> /* free, exit */
#include <ctype.h> /* toupper, tolower, isalpha */
#include <string.h> /* strlen, strpbrk */

#include "fasta.h"

#ifndef VERSION
#define VERSION ?.?
#endif

#define STRINGIFY(x) #x
#define TOSTRING(x) STRINGIFY(x)

#define DEFAULT_WIDTH 70

#define CFLAG 0x01
#define GFLAG 0x02
#define IFLAG 0x04
#define SFLAG 0x08
#define TFLAG 0x10
#define UFLAG 0x20
#define LFLAG 0x40

#define VFLAG 0x1000

int flags; /* argument flags */
int warg = DEFAULT_WIDTH; /* width argument */

int main(int argc, char *argv[])
{
	int nok = 0;
	char *s;

	while (--argc > 0 && (*++argv)[0] == '-')
		for (s = argv[0]+1; *s; s++)
			switch (*s) {
			case 'c':
				flags |= CFLAG;
				break;
			case 'g':
				flags |= GFLAG;
				break;
			case 'i':
				flags |= IFLAG;
				break;
			case 's':
				flags |= SFLAG;
				break;
			case 't':
				flags |= TFLAG;
				break;
			case 'u':
				flags |= UFLAG;
				if (flags & LFLAG) nok = 1;
				break;
			case 'l':
				flags |= LFLAG;
				if (flags & UFLAG) nok = 1;
				break;
			case 'w':
				if (isdigit(s[1])) warg = atoi(s+1);
				else nok = 1;
				s += strlen(s+1);
				break;
			case 'v':
				flags |= VFLAG;
				break;
			default:
				nok = 1;
				fprintf(stderr, " -- not an option: %c\n", *s);
				break;
			}

	if (flags & VFLAG) fprintf(stderr, "Version: %s\n", TOSTRING(VERSION));
	if (nok || argc < 1 || (argc > 0 && *argv[0] == '?')) {
		fprintf(stderr, "Usage: fac -cgistulv -w<int> <in_file> [<out_file>]\n");
		fprintf(stderr, "  -c  change ambiguous symbols to N\n");
		fprintf(stderr, "  -g  strip gap symbols\n");
		fprintf(stderr, "  -i  strip description from header, keep entry ID\n");
		fprintf(stderr, "  -s  strip entries with all ambiguous symbols\n");
		fprintf(stderr, "  -t  strip entries with any ambiguous symbols\n");
		fprintf(stderr, "  -u  change symbols to upper case\n");
		fprintf(stderr, "  -l  change symbols to lower case\n");
		fprintf(stderr, "  -w  width (columns) of entry, default %d\n", DEFAULT_WIDTH);
		fprintf(stderr, "  -v  version\n");
		exit(EXIT_FAILURE);
	}
	{
		FILE *fin, *fout;
		sequence entry;
		int len;

		if ((fin = fopen(argv[0], "r")) == NULL) {
			fprintf(stderr, " -- can't open file: %s\n", argv[0]);
			exit(EXIT_FAILURE);
		}
		if (argc < 2) {
			fout = stdout;
		} else if ((fout = fopen(argv[1], "w")) == NULL) {
			fprintf(stderr, " -- can't open file: %s\n", argv[1]);
			exit(EXIT_FAILURE);
		}
		while ((len = Fasta_Read_Entry(fin, &entry)) > 0) {
			int i, j = 0;
			int ncount = 0;
			for (i = 0; i < len; i++) {
				int c = entry.str[i];
				switch (c) {
				case 'A': case 'a':
				case 'G': case 'g':
				case 'C': case 'c':
				case 'T': case 't':
				case 'U': case 'u':
				case '-':
					break;
				default:
					if (flags & CFLAG) c = 'N';
					ncount++;
					break;
				}
				if (flags & UFLAG) c = toupper(c);
				if (flags & LFLAG) c = tolower(c);
				if (~flags & GFLAG || isalpha(c)) entry.str[j++] = c;
			}
			entry.str[j] = '\0';
			if (flags & SFLAG && ncount == len) continue;
			if (flags & TFLAG && ncount > 0) continue;
			if (flags & IFLAG) {
				char *s;
				if ((s = strpbrk(entry.hdr, " |")) != NULL) *s = '\0';
			}
			Fasta_Write_File(fout, &entry, 1, warg);
			free(entry.hdr);
			free(entry.str);
		}
		fclose(fin);
		fclose(fout);
	}
	return(EXIT_SUCCESS);
}
