/*
$Id: fas.c,v 1.4 2010/09/23 18:10:41 scott Exp $

Description: FASTA file statistics

$Log: fas.c,v $
Revision 1.4  2010/09/23 18:10:41  scott
Fix compatibility with gnuplot 4.0

Revision 1.3  2010/08/06 01:11:04  scott
Use 'stderr' for printing errors and usage.

Revision 1.2  2010/07/27 21:47:17  scott
Update for calls to Fasta_ routine

History:
 04-17-09 Scott: original revision
 11-17-09 Scott: use fasta.[ch] and path.[ch]

*/

#include <stdio.h>
#include <stdlib.h> /* calloc, malloc, realloc, free, exit */
#include <ctype.h> /* isspace, isdigit, isgraph */
#include <string.h> /* strlen, strrchr, memcpy, memset */
#include <assert.h>

#include "fasta.h"
#include "path.h"

#ifndef VERSION
#define VERSION ?.?
#endif

#define STRINGIFY(x) #x
#define TOSTRING(x) STRINGIFY(x)

#define DIRC '/' /* directory separator character */

#define RANGE_LO 20000
#define RANGE_HI 20000

#define EFLAG 0x01
#define IFLAG 0x02
#define TFLAG 0x04
#define HFLAG 0x08
#define LFLAG 0x10
#define MFLAG 0x20
#define NFLAG 0x40
#define QFLAG 0x80

#define AFLAG (EFLAG|TFLAG|MFLAG|NFLAG|QFLAG)
#define VFLAG 0x1000

int flags; /* argument flags */
int garg; /* graph argument */
char *farg; /* file argument */

/* * * * * * * * * * Binary Search Tree * * * * * * * * * */

typedef unsigned int bkey_t;
typedef unsigned int bvalue_t;

typedef struct _node {
	struct _node *left, *right;
	bkey_t key;
	bvalue_t value;
} node_t;

node_t *bst_root;
void (*bst_visit)(node_t *);

void bst_add(node_t **nptr, bkey_t key)
{
	if (*nptr == NULL) {
		*nptr = (node_t *)calloc(1, sizeof(node_t));
		(*nptr)->key = key;
		(*nptr)->value = 1;
	} else if (key < (*nptr)->key) {
		bst_add(&(*nptr)->left, key);
	} else if (key > (*nptr)->key) {
		bst_add(&(*nptr)->right, key);
	} else {
		(*nptr)->value++;
	}
}

bkey_t bst_min(node_t *nptr)
{
	if (nptr == NULL) return(0);
	while (nptr->left != NULL)
		nptr = nptr->left;
	return(nptr->key);
}

bkey_t bst_max(node_t *nptr)
{
	if (nptr == NULL) return(0);
	while (nptr->right != NULL)
		nptr = nptr->right;
	return(nptr->key);
}

void bst_inorder(node_t *nptr)
{
	if (nptr == NULL) return;
	bst_inorder(nptr->left);
	(*bst_visit)(nptr);
	bst_inorder(nptr->right);
}

/* * * * * * * * * * N50 Calculations * * * * * * * * * */

static int N50sym; /* half the total symbols */
static int N50sum = 0; /* running sum of symbols from smallest to largest length */
static int N50len = 0; /* length of sequences (and less) that total half the symbols */

static void visit_N50(node_t *nptr)
{
	N50sum += nptr->key * nptr->value;
	if (N50sum > N50sym && N50len == 0)
		N50len = nptr->key;
}

/* * * * * * * * * * Histogram Calculations * * * * * * * * * */

static double hoff; /* offset to first bin */
static double hscale; /* scale to number of bins */
static int nbin; /* number of histogram bins */
static int *hbin; /* histogram bins */

static void visit_hbin(node_t *nptr)
{
	int i = (int)((nptr->key-hoff)*hscale+0.5f);
	if (i < 0 || i >= nbin) return;
	hbin[i] += nptr->value;
}

/* * * * * * * * * * Quartile Calculations * * * * * * * * * */

static int q1, q2, q3; /* quartiles */
static int i1, i2, i3; /* index in sorted lengths for the quartiles */
static int cidx = 0; /* current index */
static int mcount = 0; /* mode count */
static int mode = 0; /* most often occurring length */

static void visit_quartile(node_t *nptr)
{
	int tidx = cidx + nptr->value;

	if (cidx <= i1 && i1 < tidx) q1 = nptr->key;
	if (cidx <= i2 && i2 < tidx) q2 = nptr->key;
	if (cidx <= i3 && i3 < tidx) q3 = nptr->key;
	cidx = tidx;
	if (nptr->value >= mcount) {
		mode = nptr->key;
		mcount = nptr->value;
	}
}

/* * * * * * * * * * ### * * * * * * * * * */


int main(int argc, char *argv[])
{
	int nok = 0;
	char *s;

	while (--argc > 0 && (*++argv)[0] == '-')
		for (s = argv[0]+1; *s; s++)
			switch (*s) {
			case 'a':
				flags |= AFLAG;
				break;
			case 'e':
				flags |= EFLAG;
				break;
			case 'i':
				flags |= IFLAG;
				break;
			case 't':
				flags |= TFLAG;
				break;
			case 'h':
				flags |= HFLAG;
				break;
			case 'l':
				flags |= LFLAG;
				break;
			case 'm':
				flags |= MFLAG;
				break;
			case 'n':
				flags |= NFLAG;
				break;
			case 'q':
				flags |= QFLAG;
				break;
			case 'g':
				if (isdigit(s[1])) garg = atoi(s+1);
				else garg = -1;
				s += strlen(s+1);
				break;
			case 'f':
				farg = s+1;
				s += strlen(s+1);
				if (garg == 0) garg = -1;
				break;
			case 'v':
				flags |= VFLAG;
				break;
			default:
				nok = 1;
				fprintf(stderr, " -- not an option: %c\n", *s);
				break;
			}

	if (flags & VFLAG) fprintf(stderr, "Version: %s\n", TOSTRING(VERSION));
	if (nok || argc < 1 || (argc > 0 && *argv[0] == '?')) {
		fprintf(stderr, "Usage: fas -aeithlmnqv -g<int> -f<out_file> <in_file>\n");
		fprintf(stderr, "  -a  all statistics, same as -etmnq\n");
		fprintf(stderr, "  -e  entry count\n");
		fprintf(stderr, "  -i  individual symbol counts\n");
		fprintf(stderr, "  -t  total symbol count\n");
		fprintf(stderr, "  -h  sequence headers\n");
		fprintf(stderr, "  -l  sequence lengths\n");
		fprintf(stderr, "  -m  min, max, and ave length\n");
		fprintf(stderr, "  -n  N50 length\n");
		fprintf(stderr, "  -q  quartiles and mode\n");
		fprintf(stderr, "  -g  graph lengths with <n> bins\n");
		fprintf(stderr, "  -f  graph to file\n");
		fprintf(stderr, "  -v  version\n");
		exit(EXIT_FAILURE);
	}
	if (!flags) flags = AFLAG;
	{
		FILE *fin;
		sequence entry;
		int ecount = 0;
		int tcount = 0;
		int minlen;
		int maxlen;
		double avelen;
		int len;
		int *icount;

		if (flags & IFLAG) icount = (int *)calloc(256, sizeof(int));
		if ((fin = fopen(argv[0], "r")) == NULL) {
			fprintf(stderr, " -- can't open file: %s\n", argv[0]);
			exit(EXIT_FAILURE);
		}
		while ((len = Fasta_Read_Entry(fin, &entry)) > 0) {
			if (flags & IFLAG) {
				int i;
				for (i = 0; i < len; i++) icount[(unsigned char)entry.str[i]]++;
			}
			if (flags & HFLAG) fprintf(stdout, " %s", entry.hdr);
			if (flags & LFLAG) fprintf(stdout, " %d", len);
			if (flags & (HFLAG|LFLAG)) fputc('\n', stdout);
			bst_add(&bst_root, len);
			ecount += 1;
			tcount += len;
			free(entry.hdr);
			free(entry.str);
		}
		fclose(fin);
		minlen = bst_min(bst_root);
		maxlen = bst_max(bst_root);
		avelen = (double)tcount/ecount;
		if (flags & EFLAG) fprintf(stdout, "sequences:%d\n", ecount);
		if (flags & TFLAG) fprintf(stdout, "symbols:%d\n", tcount);
		if (flags & MFLAG) fprintf(stdout, "min:%d max:%d ave:%.4f\n", minlen, maxlen, avelen);
		if (flags & QFLAG) {
			i1 = ecount/4;
			i2 = ecount/2;
			i3 = i1+i2;
			bst_visit = visit_quartile;
			bst_inorder(bst_root);
			fprintf(stdout, "q1:%d q2:%d q3:%d mode:%d\n", q1, q2, q3, mode);
		}
		if (flags & NFLAG) {
			N50sym = tcount/2;
			bst_visit = visit_N50;
			bst_inorder(bst_root);
			fprintf(stdout, "N50:%d\n", N50len);
		}
		if (flags & IFLAG) {
			int i;
			int sum = 0;
			for (i = 0; i < 256; i++) {
				if (icount[i]) {
					fprintf(stdout, isgraph(i) ? " %c:%d" : " #%02X:%d", i, icount[i]);
					sum += icount[i];
				}
			}
			fputc('\n', stdout);
			assert(sum == tcount);
		}
		if (garg) {
			FILE *data;
			FILE *pipe;
			char *dfile = "fas_plot.dat";
			int i, mid, range;

			// clip the range to something useful around a mid length
			mid = (flags & QFLAG) ? q2 : (int)avelen;
			if (mid-minlen > RANGE_LO) minlen = mid-RANGE_LO;
			if (maxlen-mid > RANGE_HI) maxlen = mid+RANGE_HI;
			range = maxlen - minlen;
			if (garg < 0) nbin = range+1; else nbin = garg;
			hoff = minlen;
			hscale = (double)(nbin-1) / range;
			hbin = (int *)calloc(nbin, sizeof(int));
			if (hbin == NULL) {
				fprintf(stderr, " -- error: malloc of bins\n");
				exit(EXIT_FAILURE);
			}
			bst_visit = visit_hbin;
			bst_inorder(bst_root);

			data = fopen(dfile, "w");
			for (i = 0; i < nbin; i++) fprintf(data, "%d\t%d\n", (int)((double)i/hscale + hoff), hbin[i]);
			fclose(data);
			free(hbin);

			pipe = popen("gnuplot -persist", "w"); // -persist
			if (farg != NULL) {
				char *dir, *base, *ext, *path;
				parse_path(DIRC, farg, &dir, &base, &ext);
				if (*base == '\0')
					parse_path(DIRC, argv[0], NULL, &base, NULL);
				if (*ext == '\0')
					ext = ".png";
				path = build_path(DIRC, dir, base, ext);
				fprintf(pipe, "set term %s\n", ext+1);
				fprintf(pipe, "set output '%s'\n", path);
			}
			fprintf(pipe, "set title 'Sequence Lengths for File %s'\n", argv[0]);
			fprintf(pipe, "set boxwidth 0.8 relative\n");
			fprintf(pipe, "set style fill solid 1.0\n");
			fprintf(pipe, "set xlabel 'Length'\n");
			fprintf(pipe, "set ylabel 'Count'\n");
			fprintf(pipe, "set key off\n");
			if (flags & NFLAG) {
				if (minlen <= N50len && N50len <= maxlen) {
					/* Can't use "front" in "set arrow" because of a bug in gnuplot 4.0. */
					/* Otherwise, the line is not visible. */
					fprintf(pipe, "set arrow from %d, graph 0 to %d, graph 1 nohead lt 2\n", N50len, N50len);
					fprintf(pipe, "set label ' N50=%d' at %d, graph 0.95 front\n", N50len, N50len);
				} else {
					fprintf(pipe, "set label ' N50=%d' at graph 0.8, graph 0.95 front\n", N50len);
				}
			}
			fprintf(pipe, "plot '%s' with boxes\n", dfile);
			fflush(pipe);
			//getchar();
			pclose(pipe);
			remove(dfile);
		}
	}
	return(EXIT_SUCCESS);
}
