/*
$Id: mscore.c,v 1.2 2011/12/15 21:55:00 scott Exp $

Description: Multiple Sequence Alignment scoring program

$Log: mscore.c,v $
Revision 1.2  2011/12/15 21:55:00  scott
Print usage if no scoring options are given

Revision 1.1  2010/08/04 22:21:49  scott
Initial revision

*/

#include <stdio.h> /* fprintf */
#include <stdlib.h> /* exit, atoi, malloc, calloc, free, qsort */
#include <string.h> /* strcmp */
#include <ctype.h> /* toupper */

#include "fasta.h"

#ifndef VERSION
#define VERSION ?.?
#endif

#define STRINGIFY(x) #x
#define TOSTRING(x) STRINGIFY(x)

#define min(a, b) ((a)<(b)?(a):(b))
#define max(a, b) ((a)>(b)?(a):(b))

#define IFLAG 0x01
#define QFLAG 0x02
#define TFLAG 0x04
#define CFLAG 0x08
#define FFLAG 0x10
#define VFLAG 0x1000

int flags; /* argument flags */


static int compare_entry(const void *a, const void *b)
{
	int res = strcmp(((sequence *)a)->hdr, ((sequence *)b)->hdr);
	if (res == 0) { /* string compare sequences excluding gaps */
		register const unsigned char *s = ((sequence *)a)->str;
		register const unsigned char *t = ((sequence *)b)->str;
		register unsigned char cs, ct;
		do {
			while ((cs = *s++) == '-');
			while ((ct = *t++) == '-');
			if (cs == '\0') break;
		} while (cs == ct);
		res = cs - ct;
	}
	return res;
}

static void checkup(sequence *entry, int count, int *length, char *ident)
{
	int i;

	if (count < 2) {
		fprintf(stderr, " -- must have two or more %s sequences\n", ident);
		exit(EXIT_FAILURE);
	}
	*length = entry[0].str_len;
	for (i = 0; i < count; i++) {
		register char *s, *e;
		if (entry[i].str_len != *length) {
			fprintf(stderr, " -- %s sequences differ in length\n", ident);
			exit(EXIT_FAILURE);
		}
		for (s = entry[i].str, e = s+entry[i].str_len; s < e; s++) {
			/* translate various gap symbols to '-' */
			if (*s == '.' || *s == '~') *s = '-';
			/* could also translate 'U' to 'T' */
		}
	}
}

/* verify similarity except for gaps */
static void verify_similarity(sequence *rentry, int rcount, int R_len, sequence *tentry, int tcount, int T_len)
{
	int e;

	if (rcount != tcount) {
		fprintf(stderr, " -- sequence count differs, ref:%d test:%d\n", rcount, tcount);
		exit(EXIT_FAILURE);
	}
	qsort(rentry, rcount, sizeof(sequence), compare_entry);
	qsort(tentry, tcount, sizeof(sequence), compare_entry);
	for (e = 0; e < rcount; e++) {
		char *R = rentry[e].str;
		char *T = tentry[e].str;
		int i = 0, j = 0;
		//printf("R:%s\nT:%s\n", R, T);
		while (i < R_len || j < T_len) {
			while (R[i] == '-') i++;
			while (T[j] == '-') j++;
			if (toupper(R[i]) != toupper(T[j])) {
				fprintf(stderr, " -- sequences (%d) differ, R[%d]:%c T[%d]:%c\n", e, i, R[i], j, T[j]);
				exit(EXIT_FAILURE);
			}
			i++; j++;
		}
	}
}


int main(int argc, char *argv[])
{
	int nok = 0;
	int expect_argc = 0;
	char *s;

	while (--argc > 0 && (*++argv)[0] == '-')
		for (s = argv[0]+1; *s; s++)
			switch (*s) {
			case 'c':
				flags |= CFLAG;
				break;
			case 'f':
				flags |= FFLAG;
				break;
			case 'i':
				flags |= IFLAG;
				break;
			case 'q':
				flags |= QFLAG;
				break;
			case 't':
				flags |= TFLAG;
				break;
			case 'v':
				flags |= VFLAG;
				break;
			default:
				nok = 1;
				fprintf(stderr, " -- not an option: %c\n", *s);
				break;
			}

	if (flags & IFLAG) expect_argc = 1;
	if (flags & (QFLAG|TFLAG|CFLAG|FFLAG)) expect_argc = 2;
	if (flags & VFLAG) fprintf(stderr, "Version: %s\n", TOSTRING(VERSION));
	if (nok || expect_argc == 0 || argc < expect_argc || (argc > 0 && *argv[0] == '?')) {
		fprintf(stderr, "Usage: mscore -cfiqtv <ref_file> [<test_file>]\n");
		fprintf(stderr, "  -c  column score for <ref_file> <test_file>\n");
		fprintf(stderr, "  -f  fractional identity for <ref_file> <test_file>\n");
		fprintf(stderr, "  -i  average pairwise sequence identity for <ref_file>\n");
		fprintf(stderr, "  -q  quality score for <ref_file> <test_file>\n");
		fprintf(stderr, "  -t  total column score for <ref_file> <test_file>\n");
		fprintf(stderr, "  -v  version\n");
		exit(EXIT_FAILURE);
	}

	{
		FILE *fref, *ftest;
		sequence *rentry, *tentry;
		int rcount, tcount;
		int R_len, T_len;

		/* read in alignments */
		if (expect_argc > 0) {
			if ((fref = fopen(argv[0], "r")) == NULL) {
				fprintf(stderr, " -- can't open file: %s\n", argv[0]);
				exit(EXIT_FAILURE);
			}
		}
		if (expect_argc > 1) {
			if ((ftest = fopen(argv[1], "r")) == NULL) {
				fprintf(stderr, " -- can't open file: %s\n", argv[1]);
				exit(EXIT_FAILURE);
			}
		}
		if (expect_argc > 0) {
			rcount = Fasta_Read_File(fref, &rentry);
			fclose(fref);
			checkup(rentry, rcount, &R_len, "reference");
		}
		if (expect_argc > 1) {
			tcount = Fasta_Read_File(ftest, &tentry);
			fclose(ftest);
			checkup(tentry, tcount, &T_len, "test");
			verify_similarity(rentry, rcount, R_len, tentry, tcount, T_len);
		}

/*
The Average Pairwise Sequence Identity (ASPI) is calculated using:
idents/MIN(len1,len2) for each pair of sequences where the length is
without gaps. The score reported here matches the "Average identity"
reported from "alistat" (SQUID).
*/
/*
From: SQUID - aligneval.c

Note how many ways there are to calculate pairwise identity, because of
the variety of choices for the denominator:

idents/(idents+mismat) has the disadvantage that artifactual gappy
alignments would have high "identities".

idents/(AVG|MAX)(len1,len2) both have the disadvantage that alignments
of fragments to longer sequences would have artifactually low
"identities".

Watch out in nucleic acid alignments; U/T RNA/DNA alignments will be
counted as mismatches!
*/

		if (flags & IFLAG) {
			int a, b;
			float sum = 0.0f;
			for (a = 0; a < rcount; a++) {
				char *Ra = rentry[a].str;
				for (b = 0; b < a; b++) {
					char *Rb = rentry[b].str;
					int i, len;
					int idents = 0, len1 = 0, len2 = 0;
					for (i = 0; i < R_len; i++) {
						if (Ra[i] != '-') {
							if (toupper(Ra[i]) == toupper(Rb[i])) idents++;
							len1++;
						}
						if (Rb[i] != '-') len2++;
					}
					len = min(len1, len2);
					sum += (float)idents/(float)max(1, len);
				}
			}
			fprintf(stdout, "APSI:%.3f\n", sum/(float)(rcount*(rcount-1)/2));
		}

/*
The quality (Q) score, total column (TC) score, and column score (CS)
are calculated from the reference alignment and the test alignment. The
Q and TC score are implemented as described by Carroll et al. (2007).
The column score (CS) is implemented as described in Thompson et al.
(1999b).

The Q and TC scores reported here do not match the "SP score" and "TC
score" from "bali_score" (BAliBASE). "bali_score" (V3) does not include
columns with more than 20% gaps in the "SP score" calculation. Also,
there appears to be a bug in the "TC score" calculation since it only
considers the first sequence to determine the baseline number of
columns. If sequences in an alignment are reordered, the "TC score" will
change.
*/
/*
From:

Hyrum Carroll, Wesley Beckstead, Timothy O'Connor, Mark Ebbert, Mark
Clement, Quinn Snell and David McClellan. DNA Reference Alignment
Benchmarks Based on Tertiary Structure of Encoded Proteins,
Bioinformatics Applications Note, Vol. 23 no. 19 2007, pages 26482649.

Supplementary Material:
Section 3.3 Accuracy Measurement and Statistical Analysis:

The Q score, previously termed as the developer score (Sauder et al.,
2000) or SPS (Sum of Pairs Score) (Thompson et al., 1999b), is defined
as the number of correctly aligned residue pairs in the generated
alignment divided by the number of residue pairs in the reference
alignment. The TC score, also known as the CS score, is the number of
correctly aligned columns in the generated alignment divided by the
number of columns in the reference alignment. The TC score is the same
as the Q score in the case of pairwise alignment.

References:

Sauder, J., Arthur, J., and Dunbrack Jr, R. (2000). Large-scale
comparison of protein sequence alignment algorithms with structure
alignments. Proteins Structure Function and Genetics, 40(1), 622.

Thompson, J. D., Plewniak, F., and Poch, O. (1999b). A comprehensive
comparison of multiple sequence alignment programs. Nucleic Acids
Research, 27(13), 2682 2690.
*/

		if (flags & (QFLAG|TFLAG|CFLAG)) {
			int i, a, b;
			/* for each column, count of aligned residues in... */
			int *rscore; /* ... reference */
			int *tscore; /* ... test */
			int *cscore; /* ... test, that are also aligned in reference (common) */

			int rpairs, cpairs;
			int rcolumns, ccolumns, acolumns;
			int max_col_score;

			if ((rscore = (int *)calloc(R_len, sizeof(int))) == NULL) {
				fprintf(stderr, " -- can't allocate rscore\n");
				exit(EXIT_FAILURE);
			}
			if ((tscore = (int *)calloc(T_len, sizeof(int))) == NULL) {
				fprintf(stderr, " -- can't allocate tscore\n");
				exit(EXIT_FAILURE);
			}
			if ((cscore = (int *)calloc(T_len, sizeof(int))) == NULL) {
				fprintf(stderr, " -- can't allocate cscore\n");
				exit(EXIT_FAILURE);
			}
			/* TODO: calculate the number of gaps "ngaps[i]" in each column */
			for (a = 0; a < rcount; a++) {
				char *Ra = rentry[a].str;
				char *Ta = tentry[a].str;
				for (b = 0; b < a; b++) {
					char *Rb = rentry[b].str;
					char *Tb = tentry[b].str;
					int i = 0, j = 0, rbp = 0, tbp = 0;
					for (;;) {
						int flag = 0;
						while (Ra[i] == '-') {if (Rb[i] != '-') rbp++; i++;}
						while (Ta[j] == '-') {if (Tb[j] != '-') tbp++; j++;}
						if (i >= R_len || j >= T_len) break;
						/* TODO: binc = (ngap[i] < cutoff) ? 1 : 0; */
						if (Rb[i] != '-') {rbp++; flag += 1; rscore[i]++;}
						if (Tb[j] != '-') {tbp++; flag += 1; tscore[j]++;}
						if (flag == 2 && rbp == tbp) cscore[j]++;
						i++; j++;
					}
				}
			}
			max_col_score = rcount*(rcount-1)/2;
			rcolumns = rpairs = 0;
			for (i = 0; i < R_len; i++) {
				rpairs += rscore[i];
				if (rscore[i] == max_col_score) {
					rcolumns++;
				}
			}
			acolumns = ccolumns = cpairs = 0;
			for (i = 0; i < T_len; i++) {
				cpairs += cscore[i];
				if (cscore[i] == max_col_score) {
					ccolumns++;
				}
				if (cscore[i] == tscore[i]) {
					acolumns++;
				}
			}
			free(rscore);
			free(tscore);
			free(cscore);
			// printf("cpairs:%d rpairs:%d\n", cpairs, rpairs);
			// printf("ccolumns:%d rcolumns:%d\n", ccolumns, rcolumns);
			// printf("acolumns:%d tlen:%d\n", acolumns, T_len);
			if (flags & QFLAG) fprintf(stdout, "Q:%.3f\n",
				(rpairs) ? (float)cpairs/(float)rpairs : 0.0f);
			if (flags & TFLAG) fprintf(stdout, "TC:%.3f\n",
				(rcolumns) ? (float)ccolumns/(float)rcolumns : 0.0f);
			if (flags & CFLAG) fprintf(stdout, "CS:%.3f\n",
				(T_len) ? (float)acolumns/(float)T_len : 0.0f);
		}

/*
The fractional identity (FI) is calculated from the reference alignment
and the test alignment. The FI is implemented here as described by Eddy
(1992) in the "compalign" source, although the implementation of
"compalign" is slightly different. In "compalign", non-gap symbols that
match in a pairwise comparison count double. To get the score as
reported by "compalign", #define FINC 2.
*/
/* From: SQUID main for compalign
 *
 * Compalign -- a program to compare two sequence alignments
 * SRE, Tue Nov  3 07:38:03 1992
 * RCS $Id: mscore.c,v 1.2 2011/12/15 21:55:00 scott Exp $
 *
 * incorporated into SQUID, Thu Jan 26 16:52:41 1995
 *
 * Usage: compalign <trusted-alignment> <test-alignment>
 *
 * Calculate the fractional "identity" between the trusted alignment
 * and the test alignment. The two files must contain exactly the same
 * sequences, in exactly the same order.
 *
 * The identity of the multiple sequence alignments is defined as
 * the averaged identity over all N(N-1)/2 pairwise alignments.
 *
 * The fractional identity of two sets of pairwise alignments
 * is in turn defined as follows (for aligned known sequences k1 and k2,
 * and aligned test sequences t1 and t2):
 *
 *           matched columns / total columns,
 *
 *       where total columns = the total number of columns in
 *        which there is a valid (nongap) symbol in k1 or k2;
 *
 *       matched columns = the number of columns in which one of the
 *         following is true:
 *
 *          k1 and k2 both have valid symbols at a given column; t1 and t2
 *             have the same symbols aligned in a column of the t1/t2
 *             alignment;
 *
 *          k1 has a symbol aligned to a gap in k2; that symbol in t1
 *             is also aligned to a gap;
 *
 *          k2 has a symbol aligned to a gap in k1; that symbol in t2
 *             is also aligned to a gap.
 *
 * Because scores for all possible pairs are calculated, the
 * algorithm is of order (N^2)L for N sequences of length L;
 * large sequence sets will take a while.
 *
 * Sean Eddy, Tue Nov  3 07:46:59 1992
 *
 */

		#define FINC 1
		if (flags & FFLAG) {
			int a, b;
			float sum = 0.0f;
			for (a = 0; a < rcount; a++) {
				char *Ra = rentry[a].str;
				char *Ta = tentry[a].str;
				for (b = 0; b < a; b++) {
					char *Rb = rentry[b].str;
					char *Tb = tentry[b].str;
					int mpairs = 0, tpairs = 0;
					int i = 0, ja = -1, jb = -1;
					while (i < R_len) {
						register int flaga = 0, flagb = 0;
						if (Ra[i] != '-') {
							while (Ta[++ja] == '-') ;
							flaga = 1;
						}
						if (Rb[i] != '-') {
							while (Tb[++jb] == '-') ;
							flagb = 1;
						}
						if (flaga && flagb) {
							if (ja == jb) mpairs += FINC;
							tpairs += FINC;
						} else if (flaga) {
							if (Tb[ja] == '-') mpairs++;
							tpairs++;
						} else if (flagb) {
							if (Ta[jb] == '-') mpairs++;
							tpairs++;
						}
						i++;
					}
					sum += (float)mpairs/(float)max(1, tpairs);
				}
			}
			fprintf(stdout, "FI:%.3f\n", sum/(float)(rcount*(rcount-1)/2));
		}
	}

	return(EXIT_SUCCESS); /* or EXIT_FAILURE */
}

/* TODO:
 * Implement an -a "all" option that sets all the scoring options.
 * Implement an -f1 option that sets FINC to 2.
 * Implement a -b "bali_score" option that sets the gap cutoff to 20%.
   Print BSP:0.000 and BTC:0.000.
 */
